from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.protocols import amp
from twisted.internet.protocol import ClientCreator
from twisted.internet.protocol import Factory

import libtcodpy as libtcod
import textwrap
import math

from game_protocol import *

PORT = 1337

SCREEN_WIDTH = 80
SCREEN_HEIGHT = 50
LIMIT_FPS = 20

MAP_WIDTH = 80
MAP_HEIGHT = 45

#GUI information
BAR_WIDTH = 20
PANEL_HEIGHT = 5
PANEL_Y = SCREEN_HEIGHT - PANEL_HEIGHT

#MSG bar info
MSG_X = BAR_WIDTH + 2
MSG_WIDTH = SCREEN_WIDTH - BAR_WIDTH - 2
MSG_HEIGHT = PANEL_HEIGHT - 1

#Game vars
DEFAULT_SPEED = 3
PLAYER_SPEED = 1
GUN_RANGE = 20
RECHARGE = 10
BLASTWAVE_RADIUS = 10

COLOR_LIGHT_WALL = libtcod.Color(130, 110, 50)
COLOR_DARK_WALL = libtcod.Color(0, 0, 100)

COLOR_LIGHT_GROUND = libtcod.Color(200, 180, 50)
COLOR_DARK_GROUND = libtcod.Color(50, 50, 150)

COLOR_DEAD = libtcod.Color(10, 10, 40)

TILE_COLOR = {
    # visible, wall
    (True, True): COLOR_LIGHT_WALL,
    (True, False): COLOR_LIGHT_GROUND,
    (False, True): COLOR_DARK_WALL,
    (False, False): COLOR_DARK_GROUND,
}

FOV_ALGO = 0  #default FOV algorithm
FOV_LIGHT_WALLS = True
TORCH_RADIUS = 100

MAX_ROOM_NPC = 5
MAX_MAP_NPC = 50

libtcod.console_set_custom_font(
    'arial10x10.png',
    libtcod.FONT_TYPE_GREYSCALE | libtcod.FONT_LAYOUT_TCOD)

libtcod.console_init_root(
    SCREEN_WIDTH,
    SCREEN_HEIGHT,
    'NKN', False)

## libtcod.sys_set_fps(LIMIT_FPS) SERVER DOES THIS NOW.

def dbprint(s):
    print s

class Menu(object):
    def __init__(self, header, options, width, escape_val=None, interactive=True):
        """Set escape_val to something else to return a value on escape"""
        if len(options) > 26:
            raise ValueError('Cannot have a menu with more than 26 options.')
    
        #calculate total height for the header (after auto-wrap) and one line per option
        self.header_height = libtcod.console_height_left_rect(0, 0, 0, width, SCREEN_HEIGHT, header)
        self.height = max(len(options), 10) + self.header_height
        
        #create an off-screen console that represents the menu's window
        self.window = libtcod.console_new(width, self.height)
        
        #print the header, with auto-wrap
        libtcod.console_set_foreground_color(self.window, libtcod.white)
        libtcod.console_print_left_rect(self.window, 0, 0, width, self.height, libtcod.BKGND_NONE, header)
        
        self.selected = 0 if interactive else -1
        self.header = header
        self.options = options
        self.width = width
        self.escape_val = escape_val
        self.interactive = interactive
    
    def highlight_index(self, i):
        return i == self.selected
    
    def showframe(self):
        #print all the options
        y = self.header_height
        letter_index = ord('a')
        for i, option_text in enumerate(self.options):
            if self.interactive:
                text = '(' + chr(letter_index) + ') ' + option_text
            else:
                text = option_text
            if self.highlight_index(i):
                libtcod.console_set_foreground_color(self.window, libtcod.white)
            else:
                libtcod.console_set_foreground_color(self.window, libtcod.grey)
            
            libtcod.console_print_left(self.window, 0, y, libtcod.BKGND_NONE, text)
            y += 1
            letter_index += 1
    
        #blit the contents of "window" to the root console
        x = SCREEN_WIDTH/2 - self.width/2
        y = SCREEN_HEIGHT/2 - self.height/2
        libtcod.console_blit(self.window, 0, 0, self.width, self.height, 0, x, y, 1.0, 0.7)
    
        #present the root console to the player and wait for a key-press
        libtcod.console_flush()
    
    def runframe(self):
        self.showframe()
        
        if self.interactive:
            key = libtcod.console_wait_for_keypress(True)
            if key.vk == libtcod.KEY_UP:
                self.selected = max(0, self.selected - 1)
            elif key.vk == libtcod.KEY_DOWN:
                self.selected = min(self.selected + 1, len(self.options) - 1)
            elif key.vk == libtcod.KEY_ENTER:
                return self.selected
            elif key.vk == libtcod.KEY_ESCAPE and escapable:
                return escape_val
            else:
                #convert the ASCII code to an index; if it corresponds to an option, return it
                index = key.c - ord('a')
                if index >= 0 and index < len(self.options):
                    self.selected = index
        else:
            return libtcod.console_check_for_keypress().vk == libtcod.KEY_SPACE
        
        return None

class LobbyMenu(Menu):
    def __init__(self, header):
        self.registered_players = []
        self.confirmed_players = set()
        super(LobbyMenu, self).__init__(header, self.registered_players, 60, interactive=False)
        
    def highlight_index(self, i):
        return self.registered_players[i] in self.confirmed_players

def menu(*args, **kwargs):
    m = Menu(*args, **kwargs)
    while 1:
        x = m.runframe()
        if x is not None:
            return x

class GameWorld(object):
    def __init__(self, screen_width, screen_height, level, objects, prng_seed):
        self.width = screen_width
        self.height = screen_height
        
        self.level = level
        
        self.objects = objects
        self.killed = set()
        self.recharge = RECHARGE
        self.num_npcs = 1
        self.kill_count = 0
        
        self.pillars = []
        
        self.init_fov_map() # sets self.fov_map, self.fov_recompute
        
        self.rng = libtcod.random_new_from_seed(prng_seed)
        
        self.console = libtcod.console_new(screen_width, screen_height)
        libtcod.console_set_foreground_color(self.console, libtcod.white)
        self.panel = libtcod.console_new(screen_width, PANEL_HEIGHT)
        
        self.place_map_objects()
        
        for obj in self.objects:
            obj.world = self
        
    
    def init_fov_map(self):
        self.fov_map = fov_map = libtcod.map_new(MAP_WIDTH, MAP_HEIGHT)
        self.fov_recompute = True
        
        for x, tiles in enumerate(self.level):
            for y, tile in enumerate(tiles):
                libtcod.map_set_properties(
                    fov_map, 
                    x,
                    y, 
                    not tile.block_sight,
                    not tile.blocked)
    
    def read_map(self, filepath):
        f = open(filepath, 'r')
        line = f.readline()
        y = 0
        while line != '':
            rough_map[y] = line
            y += 1
            line = f.readline()
        f.close()
        MAP_HEIGHT = y - 1
        x = len(rough_map[0]) - 1
        if rough_map[0][x] == '\n':
            x -= 1
        i = k = 0
        while (k <= y):
            while (i <= x):
                if rough_map[k][i] != ' ':
                    self.pillars.append((i, k))
    
    def find_random_pos(self):
        while 1:
            # hypothetically could loop forever -- doubt it.
            
            x = libtcod.random_get_int(self.rng, 0, MAP_WIDTH - 2)
            y = libtcod.random_get_int(self.rng, 0, MAP_HEIGHT - 2)
            
            if not self.is_blocked(x, y):
                return x, y
    
    def place_map_objects(self):
        #choose random number of npcs
        self.num_npcs = libtcod.random_get_int(self.rng, 20, MAX_MAP_NPC)
    
        for i in range(self.num_npcs):
            #choose random spot for this npc
            x, y = self.find_random_pos()
            
            if libtcod.random_get_int(self.rng, 0, 100) < 80:  #80% chance of getting an orc
                #create an orc
                npc = Mob(x, y, 'o', 'Orc', libtcod.desaturated_green, blocking=True)
            else:
                #create a troll
                npc = Mob(x, y, 'T', 'TROLL', libtcod.darker_green, blocking=True)

            self.objects.append(npc)
                
    def place_object(self, obj):
        self.objects.append(obj)
    
    def draw_map(self):
        for x, tiles in enumerate(self.level):
            for y, tile in enumerate(tiles):
                visible = libtcod.map_is_in_fov(self.fov_map, x, y)
                wall = tile.block_sight
                
                color = TILE_COLOR[visible, wall]
                
                if visible:
                    tile.explored = True
                if tile.explored:
                    libtcod.console_set_back(self.console,
                        x, y, color, libtcod.BKGND_SET)
    
    def draw_obj(self, obj):
        # set the color and then draw the character that represents this object
        # at its position
        if not libtcod.map_is_in_fov(self.fov_map, obj.x, obj.y):
            return
        
        libtcod.console_set_foreground_color(self.console, obj.color)
        libtcod.console_put_char(self.console,
            obj.x, obj.y, obj.char, libtcod.BKGND_NONE)

    def draw_all(self):
        if self.fov_recompute:
            #recompute FOV if needed (the player moved or something)
            self.fov_recompute = False
            libtcod.map_compute_fov(
                self.fov_map,
                self.player_obj.x,
                self.player_obj.y,
                TORCH_RADIUS,
                FOV_LIGHT_WALLS,
                FOV_ALGO)
        self.draw_map()
        
        
        for obj in self.killed:
            self.draw_obj(obj)

        #Gunshot grapics
        for obj in self.objects:
            if obj.time_left == 0:
                obj.char = ' '
                self.draw_obj(obj)
                self.objects.remove(obj)
                
        for obj in self.objects:
            if obj != self.player_obj and obj not in self.killed:
                self.draw_obj(obj)
                
        self.draw_obj(self.player_obj)
        
        self.blit()
        
        #prepare to render the GUI panel
        libtcod.console_set_background_color(self.panel, libtcod.black)
        libtcod.console_clear(self.panel)
        
        y = 1
        for (line, color) in game_msgs:
            libtcod.console_set_foreground_color(self.panel, color)
            libtcod.console_print_left(self.panel, MSG_X, y, libtcod.BKGND_NONE, line)
            y += 1
 
        #shows the player recharge time on gun
        render_bar(1, 1, BAR_WIDTH, 'Recharge',self.recharge, RECHARGE,
            libtcod.light_orange, libtcod.darker_orange, self.panel)
        render_bar(1, 3, BAR_WIDTH, 'Kills', self.kill_count, self.num_npcs,
            libtcod.light_red, libtcod.darker_red, self.panel)
 
        #blit the contents of "panel" to the root console
        libtcod.console_blit(self.panel, 0, 0, SCREEN_WIDTH, PANEL_HEIGHT, 0, 0, PANEL_Y)

    def clear_obj(self, obj):
        libtcod.console_put_char(self.console,
            obj.x, obj.y, ' ', libtcod.BKGND_NONE)
    
    def clear_all_objs(self):
        for obj in self.objects:
            self.clear_obj(obj)
    
    def blit(self): # blit to screen
        # no idea what this means
        libtcod.console_blit(self.console,
            0, 0, self.width, self.height, 0, 0, 0)

    def get_mob_indices(self, x, y):
        """
        Return list indices of monsters at x, y.
        If there are none, return empty list.
        """
        # (A list, in case we ever decide to have more than one monster at a
        #  location)
        #  
        return [
            i for i, obj in enumerate(self.objects)
            if (obj.x, obj.y) == (x, y)
                and obj.blocking and obj.alive]

    def is_blocked(self, x, y):
        return self.level[x][y].blocked or bool(self.get_mob_indices(x, y))

    def is_valid_move(self, x, y):
        if not self.is_blocked(x, y) and (MAP_WIDTH - 1) > x > -1 and (MAP_HEIGHT - 1) > y > -1:
                return True
        else:
            False

    def play_effect_frame(self):
        pass

class Tile(object):
    """A tile on the map"""
    def __init__(self, blocked, block_sight = None):
        if block_sight is None:
            block_sight = blocked
        
        self.blocked = blocked
        self.block_sight = block_sight
        self.explored = False

def make_map(width, height):
    return [[Tile(False)
                for y in xrange(height)]
                    for x in xrange(width)]

def message(new_msg, color=libtcod.white):
    new_msg_lines = textwrap.wrap(new_msg, MSG_WIDTH)
    for line in new_msg_lines:
        if len(game_msgs) == MSG_HEIGHT:
            del game_msgs[0]
        game_msgs.append((line, color))


def render_bar(x, y, total_width, name, value, maximum, bar_color, back_color, panel):
    #Render a bar (HP, experience, etc).
    bar_width = int(float(value) / maximum * total_width)
 
    #Render the background
    libtcod.console_set_background_color(panel, back_color)
    libtcod.console_rect(panel, x, y, total_width, 1, False)
 
    #Render the bar on top
    libtcod.console_set_background_color(panel, bar_color)
    if bar_width > 0:
        libtcod.console_rect(panel, x, y, bar_width, 1, False)
        
    #Centered text with the values
    libtcod.console_set_foreground_color(panel, libtcod.red)
    libtcod.console_print_center(panel, x + total_width / 2, y, libtcod.BKGND_NONE,
        name + ': ' + str(value) + '/' + str(maximum))
        
class Object(object):
    """A single-time object that can be drawn onto a buffer."""
    def __init__(self, x, y, char, name, color, blocking=True, alive=True, speed=DEFAULT_SPEED, world=None):
        self.x = x
        self.y = y
        self.char = char
        self.color = color
        self.blocking = blocking
        self.name = name
        self.alive = alive
        self.speed = speed
        self.wait = 0
        self.world = world
        
        #Special attribute for timed removal
        self.time_left = 10
        
        #Special attribute for gun killing
        self.kill_coords = []
    
    def move_to(self, x, y):
        #move to the given coordinates
        if not self.alive:
            return
        self.x = x
        self.y = y
        self.wait = self.speed
        
    def distance_from(self, x, y):
        #return the distance to some coordinates
        return math.sqrt((x - self.x) ** 2 + (y - self.y) ** 2)
    
    def take_turn(self):
        pass

class Mob(Object):
    def kill(self):
        self.blocking = False
        self.color = COLOR_DEAD
        self.alive = False
        message(self.name + ' has died.', libtcod.red)
    
    def take_turn(self):
        if self.wait > 0:
            self.wait -= 1
        else:
            self.random_move()
        
    def move_obj(self, x, y):
        """Move object to x, y iff you can."""
        if self.world.is_valid_move(x, y):
            self.move_to(x, y)
            return True
        return False

    def random_move(self):
        if self.alive:
            valx = libtcod.random_get_int(self.world.rng, -1, 1)
            valy = libtcod.random_get_int(self.world.rng, -1, 1)
            tx = self.x + valx
            ty = self.y + valy
            if self.world.is_valid_move(tx, ty):
                self.move_to(tx, ty)
            
class Pew(Object):
    '''A bullet for a gun'''
    def take_turn(self):
        if self.time_left > 0:
            self.time_left -= 1

class PlayerObject(Mob):
    def take_turn(self):
        pass
    
    def kill(self):
        self.blocking = False
        self.color = COLOR_DEAD
        self.alive = False
        message(self.name + ' has died.', libtcod.red)

    def fire_gun(self, tarx, tary):
        self.kill_coords = []
        message(self.name + ' shoots his gun.', libtcod.orange)
        libtcod.line_init(self.x, self.y, tarx, tary)
        self.draw_line()
        self.world.recharge = 0
        return [
            i for i, obj in enumerate(self.world.objects)
            if  obj.alive and (obj.x, obj.y) in self.kill_coords and obj.name != 'pew' and obj != self]

    def draw_line(self):
        (x, y) = libtcod.line_step()
        if (x, y) != (None, None):
            pew = Pew(x, y, '*', 'pew', libtcod.silver, blocking=False, alive=False)
            self.world.place_object(pew)
            self.kill_coords.append((x, y))
            self.draw_line()
        
    def cast_blastwave(self, tarx, tary):
        message('An inferno rains down from the sky, incinerating everything within ' + str(BLASTWAVE_RADIUS) + ' tiles!', libtcod.orange)
        
        return [
            i for i, obj in enumerate(self.world.objects)
            if  obj.alive
                and obj.distance_from(tarx, tary) <= BLASTWAVE_RADIUS
                and obj != self]
        
        
    def player_move(self, x, y):
        if self.world.recharge < RECHARGE:
            self.world.recharge += 1
        if self.wait > 0:
            self.wait -= 1
        else:
            oldx = self.x
            oldy = self.y
            if oldx != x or oldy != y:
                self.move_obj(x, y)
                # FOV has changed
                self.world.fov_recompute = True

def place_pillars(level):
    level[30][22].blocked = True
    level[30][22].block_sight = True
    level[50][22].blocked = True
    level[50][22].block_sight = True
    level[30][21].blocked = True
    level[30][21].block_sight = True
    level[30][23].blocked = True
    level[30][23].block_sight = True
    level[50][21].blocked = True
    level[50][21].block_sight = True
    level[50][23].blocked = True
    level[50][23].block_sight = True

game_msgs = []

def forstate(state):
    def decorator(f):
        def wrapped(self, *args, **kwargs):
            if self.game_state != state:
                raise Exception(
                    "Programmer error: "
                    "HOW WE GET THIS STATE HERE: "
                    "%r [expected %r] in %s" % (self.game_state, state, wrapped.__name__))
            return f(self, *args, **kwargs)
        
        wrapped.__name__ = f.__name__
        return wrapped
    return decorator

class GameClient(amp.AMP):
    def __init__(self):
        self.game_state = 'lobby'
        self.lobby_screen = LobbyMenu("Press Space to play. The game begins when everyone consents.")
        self.lobby_ready = False
        
        # just during the lobby state
        self.looping_call = LoopingCall(self.check_lobby)
        self.looping_call.start(1.0/LIMIT_FPS)
        
        self.usernames = []
        ##self.npc = Mob(MAP_WIDTH/2 - 5, MAP_HEIGHT/2, '@', 'Jack', libtcod.yellow, blocking=True)
        self.objects = []
        ##self.objects.append(self.npc)
        
        self.level = make_map(MAP_WIDTH, MAP_HEIGHT)
        
        
        place_pillars(self.level)
        
        self.world = None
        # if still None, activity was done before Genesis event.
        
        message('Welcome!')
        
        self.panel = libtcod.console_new(SCREEN_WIDTH, PANEL_HEIGHT)
    
    def prepend_players(self, player_names, player_id):
        new_items = []
        for i, name in enumerate(player_names):
            # FIXME: randomly assign position
            x, y = self.world.find_random_pos()
            pl = PlayerObject(
                x, y, '@', name,
                libtcod.white, blocking=True, speed=PLAYER_SPEED,
                world=self.world)
            
            new_items.append(pl)
            if i == player_id:
                self.player = pl
        
        self.objects[0:0] = new_items
    
    # lobby state stuff
    @forstate('lobby')
    def check_lobby(self):
        if self.lobby_ready:
            self.lobby_screen.showframe()
        elif self.lobby_screen.runframe():
            # we are ready
            self.lobby_ready = True
            self.callRemote(IAmReady)
    
    def connectionMade(self):
        pass
        ##print "client understands connection!"
    
    @Login.responder
    @forstate('lobby')
    def hi_user(self, username):
        """Oh, a new user logged in!"""
        self.lobby_screen.registered_players.append(username)
        
        return {}
    
    @UserReady.responder
    @forstate('lobby')
    def user_ready(self, username):
        """User accepted new game"""
        self.lobby_screen.confirmed_players.add(username)
        
        return {}

    @Genesis.responder
    @forstate('lobby')
    def create_world(self, prng_seed, player_names, player_id):
        self.world = GameWorld(MAP_WIDTH, MAP_HEIGHT,
            self.level,
            self.objects,
            prng_seed=prng_seed)
        
        # mutates self.objects / world.objects
        self.prepend_players(player_names, player_id)
        self.world.player_obj = self.player
        
        
        self.game_state = 'play'
        
        self.looping_call.stop()
        
        return {}

    def handle_keys(self):
        player_obj = self.world.player_obj
        #movement keys
        dx = (libtcod.console_is_key_pressed(libtcod.KEY_RIGHT)
            - libtcod.console_is_key_pressed(libtcod.KEY_LEFT))
        dy = (libtcod.console_is_key_pressed(libtcod.KEY_DOWN)
            - libtcod.console_is_key_pressed(libtcod.KEY_UP))
        
        x, y = new_pos = player_obj.x + dx, player_obj.y + dy
        
        conflicts = self.world.get_mob_indices(x, y)
        player_i = self.objects.index(player_obj)
        if player_i in conflicts:
            conflicts.remove(player_i)
        
        if conflicts:
            # tell server to kill every id in conflicts
            self.callRemote(Kill, kill_ids=conflicts)
        else:
            self.callRemote(TryMove, char_id=player_i, new_x=x, new_y=y)
        
        #may need to put this part first
        mouse = libtcod.mouse_get_status()
        (tarx, tary) = (mouse.cx, mouse.cy)
        #Gun
        if (mouse.lbutton_pressed and libtcod.map_is_in_fov(self.world.fov_map, tarx, tary) and 
            self.world.recharge >= RECHARGE and player_obj.distance_from(tarx, tary) <= GUN_RANGE):
            self.callRemote(Kill, kill_ids=player_obj.fire_gun(tarx, tary))
        #Blastwave
        if mouse.rbutton_pressed and libtcod.console_is_key_pressed(libtcod.KEY_CONTROL):
            self.callRemote(Kill, kill_ids=player_obj.cast_blastwave(tarx, tary))
        
        key = libtcod.console_check_for_keypress()
        if key.vk == libtcod.KEY_ENTER and key.lalt:
            #Alt+Enter: toggle fullscreen
            libtcod.console_set_fullscreen(not libtcod.console_is_fullscreen())
        
        elif key.vk == libtcod.KEY_ESCAPE:
            return 'exit'  #exit game
    
    def runframe(self):
        self.world.draw_all()
        
        self.world.play_effect_frame()
        # clear the old positions
        self.world.clear_all_objs()
        #handle keys and exit game if needed
        exit = self.handle_keys()
        libtcod.console_flush()
        
        for obj in self.world.objects:
            obj.take_turn()
        
        return exit
    
    @Update.responder
    @forstate('play')
    def update_world(self, kill_ids, move_ids):
        kill_ids = set(kill_ids)
        move_ids = dict((id, (x, y)) for id, x, y in move_ids)
        
        for i, obj in enumerate(self.world.objects):
            if i in kill_ids:
                obj.kill()
                self.world.killed.add(obj)
                self.world.kill_count += 1
            if i in move_ids:
                x, y = move_ids[i]
                if obj == self.world.player_obj:
                    obj.player_move(x, y)
                    self.world.fov_recompute = True
                else:
                    obj.move_to(x, y)
        
        self.runframe()
        return {}

class GameServerConnection(amp.AMP):
    def __init__(self, server):
        self.server = server
    
    @Login.responder
    def login(self, username):
        self.username = username
        for conn in self.server.connections:
            if conn is not self:
                # update with everyone that joined earlier
                self.callRemote(Login, username=conn.username)
        
        self.server.send_login(username)
        return {}
    
    @IAmReady.responder
    def ready(self):
        self.server.send_ready(self.username)
        return {}
    
    @TryMove.responder
    def try_move(self, char_id, new_x, new_y):
        ##char = self.server.world.objects[i]
        ##char.x, char.y = new_x, new_y
        self.server.moves.add((char_id, new_x, new_y))
        return {}
    
    @Kill.responder
    def kill(self, kill_ids):
        self.server.kills.update(kill_ids)
        return {}
    
    def connectionMade(self):
        pass
    
    def connectionLost(self, reason):
        self.server.remove_conn(self, reason)

class GameServer(Factory):
    """
    Actually, client actions are just routed through here; server does no
    computation.
    
    Actually, that's a lie, TODO: is that AI is all server-side.
    
    
    :) cheating is easy (:
    
    Side effects of this decision:
    
    - clients must handle collisions between players after the fact.
      they can both decide to move to square "X, Y", which is also where the
      server decided to place them. In such a case they must both instead
      stay put.
    - cheating is really really really easy.
    
    """
    protocol = GameServerConnection
    
    def __init__(self):
        self.connections = set()
        self.players = set()
        self.moves = set() # this gets populated by connections, then cleared.
        self.kills = set()
        
        self.unready_users = set()
    
    def buildProtocol(self, addr):
        ##self.players.add(new_player)
        conn = self.protocol(self)#, player)
        self.connections.add(conn)
        
        return conn
    
    def remove_conn(self, conn, reason):
        # TODO: last man standing wins!
        self.connections.remove(conn)
        # FIXME: kill player
        # TODO: send quit message to all!
    
    def runframe(self):
        update = self.create_update()
        for conn in self.connections:
            conn.callRemote(Update, **update)
    
    def create_update(self):
        moves = self.moves
        self.moves = set()
        
        kills = self.kills
        self.kills = set()
        
        return dict(
            kill_ids = kills,
            move_ids = moves)

    def send_login(self, username):
        for conn in self.connections:
            conn.callRemote(Login, username=username)
        
        self.unready_users.add(username)
    
    def send_ready(self, username):
        for conn in self.connections:
            conn.callRemote(UserReady, username=username)
        
        self.unready_users.remove(username)
        if not self.unready_users:
            # everyone's ready
            self.start_game()

    def start_game(self):
        for i, conn in enumerate(self.connections):
            # FIXME: race condition?
            # Looks like genesis gets processed after below looping call
            conn.callRemote(Genesis,
                prng_seed=42,
                player_names=[conn.username for conn in self.connections],
                player_id=i)
        
        self.looping_call = loop = LoopingCall(self.runframe)
        self.looping_call.start(1.0/LIMIT_FPS)

    _next_id = 1
    def obj_id(self):
        i = self._next_id
        self._next_id += 1

##def main():
    ##game = Game()
    ##def loop_f():
        ##if game.runframe():
            ##loop.stop()
            ##reactor.stop()
    
    ##loop = LoopingCall(loop_f)
    ##loop.start(1/60.)find
    
    ##reactor.run()

##def servermain():
    ##game = Game()
    ##fact = Factory()
    ##fact.protocol = GameServer
    ##reactor.listenTCP(1337, fact)
    ###npc movement
    ##for obj in self.objects:
        ##obj.take_turn()

def enter_field(question):
    name = ''
    while 1:
        libtcod.console_clear(0)
        libtcod.console_print_center(0, MAP_WIDTH//2, MAP_HEIGHT//2 - 1, libtcod.BKGND_NONE, question)
        libtcod.console_print_center(0, MAP_WIDTH//2, MAP_HEIGHT//2+ 1, libtcod.BKGND_NONE, name)
        libtcod.console_flush()
        key = libtcod.console_wait_for_keypress(True)
        c = chr(key.c)
        if key.vk == libtcod.KEY_BACKSPACE:
            name = name[:-1]
        elif key.vk == libtcod.KEY_ENTER:
            libtcod.console_clear(0)
            return name
        elif (c.isalnum() or c in '._-*~') and len(name) < 20:
            name += c

def create_game_menu():
    """Player selected 'create new game'.
    
    Basically just means spin up a server, wait for others to connect.
        
    """
    name = enter_field("Please type in your username:")
    server = GameServer()
    reactor.listenTCP(PORT, server)
    
    ClientCreator(reactor,GameClient).connectTCP('localhost', PORT).addCallback(
        lambda p: p.callRemote(Login, username=name))
    
    reactor.run()

def join_game_menu():
    host = enter_field("Please type in the internet address of the server:")
    ClientCreator(reactor,GameClient).connectTCP('localhost', PORT).addCallback(
        lambda p: p.callRemote(Login, username=enter_field("Please type in your username:")))
    
    reactor.run()


def main():
    ##img = libtcod.image_load('menu_background.png')
 
    console = libtcod.console_new(MAP_WIDTH, MAP_HEIGHT)
 
    while not libtcod.console_is_window_closed():
        #show the background image, at twice the regular console resolution
        ##libtcod.image_blit_2x(img, 0, 0, 0)
 
        #show options and wait for the player's choice
        choice = menu(
            'Welcome to NKN', [
                'Create game',
                'Join game',
                'Quit'], 24,
            escape_val=2)
 
        if choice == 0:  #new game
            return create_game_menu()
        elif choice == 1:
            return join_game_menu()
        elif choice == 2:  #quit
            return

##sp_main()

main()